# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.classes.scan_paths"""

import copy
import gc
import hashlib
import json
import logging
import os
import tarfile
import threading
import time
import uuid

from datetime import datetime

import bitmath
import six
from six.moves import queue as _queue

from boto3.s3 import transfer as S3Transfer

from dwho.classes import inotify
from dwho.classes.inoplugs import DWhoInoEventPlugBase, INOPLUGS
from sonicprobe import helpers
from sonicprobe.libs import workerpool

from .common import (DATE_FORMAT, STATUS_PROCESSING, STATUS_STOPPED)
from .s3 import (XtrargonTxProgress, XTRARGON_FILE)


LOG                         = logging.getLogger('xtrargon_ng.classes.scan_paths')

DEFAULT_MAX_WORKERS         = 'auto'
DEFAULT_MAX_LIFETIME        = 3600
DEFAULT_MAX_TASKS           = 100
DEFAULT_MAX_ATTEMPTS        = 10
DEFAULT_OBJECT_TTL          = 604800
DEFAULT_DEST_FORMAT         = "/{server_id}/{backup_type}/{time:%Y-%m-%dT%H:%M:%S}"
DEFAULT_STATE_FILE_FORMAT   = DEFAULT_DEST_FORMAT + "/xtrargon.state"
DEFAULT_COMP_MAX_TOTAL_SIZE = '1G'
DEFAULT_COMP_MAX_FILE_SIZE  = '1G'
CHK_FILE_FIRST_BYTES        = {'ibd': 36}


class BackupsSyncInoPlug(DWhoInoEventPlugBase):
    PLUGIN_NAME = 'sync'

    def __init__(self):
        super(BackupsSyncInoPlug, self).__init__()
        self._files = None
        self._lock  = threading.Lock()

    def run(self, cfg_path, event, filepath):
        with self._lock:
            if self._files is None:
                return

        if not os.path.isfile(filepath):
            return

        path_options = self._get_path_all_options()
        toppath = path_options['toppath']
        #topparams = path_options['topparams']
        #topvars = path_options['topvars']

        with self._lock:
            if toppath not in self._files:
                self._files[toppath] = set()
            self._files[toppath].add(filepath)

    def get_files(self, toppath = None):
        with self._lock:
            if not toppath or not self._files:
                return self._files

            if toppath in self._files:
                return self._files[toppath]

        return False

    def at_start(self):
        with self._lock:
            if self._files is None:
                self._files = {}

    def at_stop(self):
        with self._lock:
            self._files = None


class BackupsScanPaths(threading.Thread):
    def __init__(self, module):
        threading.Thread.__init__(self)
        self.name              = 'scan'
        self._module           = module
        self._config           = module.config
        self._s3helper         = module.s3helper
        self._inotify          = None

        self._phase             = 1
        self._killed           = False

        self._object_ttl       = DEFAULT_OBJECT_TTL
        self._file_sync        = module.file_sync

        self._backup_type      = 'full'

        self._wcompress        = None
        self._wupload          = None

        self._files_cached     = {}
        self._lock_fcached     = threading.Lock()
        self._files_compressed = []
        self._files_uploaded   = {}
        self._lock_fuploaded   = threading.Lock()
        self._files_errors     = {}
        self._lock_ferrors     = threading.Lock()

        self._s3conf           = {'bucket': self._s3helper.get_bucket(),
                                  'client': self._s3helper.boto3_client(),
                                  'storage_class': self._s3helper.get_storage_class(),
                                  'tags': self._s3helper.get_tags()}

        if 'object_ttl' in self._config['backups']:
            self._object_ttl = int(self._config['backups']['object_ttl'])

        for x in ('compress', 'upload'):
            if not isinstance(self._config['backups'].get(x), dict):
                self._config['backups'][x] = {}

    @staticmethod
    def _load_workers(name, config, qtasks = None):
        max_workers     = config.get('max_workers') or DEFAULT_MAX_WORKERS
        worker_lifetime = int(config.get('worker_lifetime', DEFAULT_MAX_LIFETIME))
        max_tasks       = int(config.get('max_tasks', DEFAULT_MAX_TASKS))

        return workerpool.WorkerPool(max_workers = max_workers,
                                     queue       = qtasks,
                                     life_time   = worker_lifetime,
                                     max_tasks   = max_tasks,
                                     name        = name,
                                     auto_gc     = False)

    @staticmethod
    def _mk_sum(filepath):
        if not os.path.exists(filepath):
            return ''

        ext   = os.path.splitext(filepath)[1].strip('.')
        fstat = os.stat(filepath)
        xsum  = [str(fstat.st_size),
                 str(fstat.st_mtime),
                 '']

        if ext in CHK_FILE_FIRST_BYTES:
            with open(filepath, 'rb') as f:
                xsum[2] = hashlib.md5(f.read(CHK_FILE_FIRST_BYTES[ext])).hexdigest()

        return '|'.join(xsum)

    def _mk_filekey(self, filepath, xid = None):
        if xid:
            return "%s:%s:%s" % (self._module.server_id, xid, filepath)

        return "%s:%s" % (self._module.server_id, filepath)

    def _chk_file_cached(self, toppath, srcpath, fkey, nsum):
        with self._lock_fcached:
            if toppath not in self._files_cached:
                self._files_cached[toppath] = {}

            if srcpath in self._files_cached[toppath]:
                osum = self._files_cached[toppath][srcpath]
                if osum and osum == nsum:
                    return True

        osum = self._file_sync.get_key(fkey)
        if osum and osum == nsum:
            return True

        return False

    def _compress_dir_params(self, params):
        r = {'dest_params': None,
             'enabled': False,
             'workers_upload': True,
             'max_file_size': bitmath.parse_string_unsafe(DEFAULT_COMP_MAX_FILE_SIZE).to_Byte().value,
             'max_attempts': DEFAULT_MAX_ATTEMPTS,
             'max_total_size': bitmath.parse_string_unsafe(DEFAULT_COMP_MAX_TOTAL_SIZE).to_Byte().value}

        if not params or not isinstance(params, dict):
            return r

        r['dest_params'] = params.get('dest_params')
        r['enabled'] = params.get('enabled') or False
        r['workers_upload'] = bool(params.get('workers_upload', True))
        r['max_file_size'] = bitmath.parse_string_unsafe(
            params.get('max_file_size', DEFAULT_COMP_MAX_FILE_SIZE)).to_Byte().value
        r['max_attempts'] = self._get_max_attempts(params)
        r['max_total_size'] = bitmath.parse_string_unsafe(
            params.get('max_total_size', DEFAULT_COMP_MAX_TOTAL_SIZE)).to_Byte().value

        return r

    @staticmethod
    def _second_phase_params(params):
        r = {'enabled': True,
             'walk_from': None,
             'max_attempts': None}

        if not params or not isinstance(params, dict):
            return r

        r['enabled'] = bool(params.get('enabled', True))
        r['max_attempts'] = params.get('max_attempts')

        if params.get('walk_from') in ('cache', 'inotify'):
            r['walk_from'] = params['walk_from']

        return r

    def _s3up(self, src, dst, max_attempts = 1, delete = False):
        if not dst:
            dst = src

        if not os.path.exists(src):
            return (None, src, dst, delete)

        if self._killed:
            return (GeneratorExit(), src, dst, delete)

        last_exception = None
        attempts_delay = None

        if max_attempts > 1:
            attempts_delay = int(max_attempts / 10.0)

        for i in range(1, max_attempts + 1):
            if self._killed:
                return (GeneratorExit(), src, dst, delete)

            try:
                with S3Transfer.S3Transfer(self._s3helper.boto3_client(),
                                           self._s3helper.transfer_config()) as transfer:
                    transfer.upload_file(
                        filename   = src,
                        bucket     = self._s3conf['bucket'],
                        key        = dst.lstrip('/'),
                        callback   = XtrargonTxProgress(src, False, self._killed),
                        extra_args = {'StorageClass': self._s3conf['storage_class'],
                                      'Tagging': self._s3conf['tags']})

                LOG.info("s3 upload (src: %r, dst: %r, bucket: %r, attempt: %s/%s)",
                         src,
                         dst,
                         self._s3conf['bucket'],
                         i,
                         max_attempts)
                return (True, src, dst, delete)
            except (GeneratorExit, KeyboardInterrupt):
                return (GeneratorExit(), src, dst, delete)
            except Exception as e:
                LOG.warning("Retrying exception caught (%s), "
                            "retrying request, (attempt %s/%s)",
                            e,
                            i,
                            max_attempts)
                last_exception = e

                if attempts_delay and (i % attempts_delay) == 0:
                    time.sleep(1)

                continue

        if not isinstance(last_exception, Exception):
            last_exception = Exception(last_exception)

        if src and self._s3helper.check_ignore_error_file(src):
            LOG.info("ignore upload error on file: %r", src)
            last_exception = True

        return (last_exception, src, dst, delete)

    def _s3up_complete(self, ret):
        result, srcpath, destpath, delete = ret

        if isinstance(result, GeneratorExit):
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        elif isinstance(result, Exception):
            if destpath:
                with self._lock_ferrors:
                    self._files_errors[destpath] = repr(result)

                with self._lock_fuploaded:
                    if destpath in self._files_uploaded \
                       and self._files_uploaded[destpath].get('wait'):
                        self._files_uploaded.pop(destpath, None)
            else:
                self._module.add_errors(repr(result))
                LOG.exception(result)
        elif result is True:
            if destpath:
                with self._lock_ferrors:
                    self._files_errors.pop(destpath, None)

                with self._lock_fuploaded:
                    self._files_uploaded.pop(destpath, None)

            if delete and os.path.exists(srcpath):
                try:
                    os.unlink(srcpath)
                except Exception:
                    pass

    def _s3api(self, cmd, key = None, **kwargs):
        kwargs['Bucket'] = self._s3conf['bucket']
        if key:
            kwargs['Key'] = key.lstrip('/')
        kwargs['StorageClass'] = self._s3conf['storage_class']
        kwargs['Tagging'] = self._s3conf['tags']

        LOG.debug("s3 api: %r: %r", cmd, kwargs)
        getattr(self._s3conf['client'], cmd)(**kwargs)

    def _put_file(self, xpath, body = ''):
        self._s3api('put_object', xpath, **{'Body': body})

    @staticmethod
    def _destpath(toppath, params, srcpath, xvars):
        destpath = srcpath
        fkwargs  = xvars.copy()
        fkwargs['filepath'] = srcpath
        fkwargs['dirname']  = os.path.dirname(srcpath)
        fkwargs['filename'] = os.path.basename(srcpath)

        if params is True:
            return destpath

        if not params or not isinstance(params, dict):
            params = {'format': DEFAULT_DEST_FORMAT,
                      'prefix': None,
                      'subst':  True}

        if params.get('format'):
            destpath = params['format'].format(**fkwargs)

        if params.get('subst'):
            if isinstance(params['subst'], six.string_types):
                xsubst = params['subst']
            else:
                xsubst = toppath

            if srcpath.startswith(xsubst):
                substpath = srcpath[len(xsubst):].strip(os.path.sep)
                if params.get('format'):
                    destpath = os.path.join(os.path.sep, destpath.strip(os.path.sep), substpath)
                else:
                    destpath = os.path.join(os.path.sep, substpath)

        if params.get('prefix'):
            destpath = os.path.join(params['prefix'].strip(os.path.sep),
                                    destpath.strip(os.path.sep))

        return os.path.join(os.path.sep, destpath.strip(os.path.sep))

    @staticmethod
    def _get_max_attempts(params):
        r = int(params.get('max_attempts', DEFAULT_MAX_ATTEMPTS) or 1)
        if r < 1:
            r = 1

        return r

    @staticmethod
    def _config_inotify_workers(config = None):
        if not isinstance(config, dict):
            config = {}

        config['max_workers']     = config.get('max_workers') or DEFAULT_MAX_WORKERS
        config['worker_lifetime'] = int(config.get('worker_lifetime', DEFAULT_MAX_LIFETIME))

        return config

    @staticmethod
    def _config_inotify_path(toppath, conf):
        conf['paths'][toppath]['toppath'] = toppath

        """
        conf['paths'][toppath]['topparams'] = copy.deepcopy(params)
        conf['paths'][toppath]['topvars'] = copy.deepcopy(xvars)

        plug_name = BackupsSyncInoPlug.PLUGIN_NAME
        if plug_name not in conf['paths'][toppath]['plugins']:
            conf['paths'][toppath]['plugins'][plug_name] = True
        """

    def _load_inotify(self, conf_orig):
        if not self._inotify:
            self._inotify = inotify.DWhoInotify()

        plug_name = BackupsSyncInoPlug.PLUGIN_NAME
        if plug_name not in INOPLUGS:
            INOPLUGS.register(BackupsSyncInoPlug())

        conf = {'general': copy.deepcopy(self._config['general'])}
        conf['inotify'] = inotify.DWhoInotifyConfig()(self._inotify,
                                                      copy.deepcopy(
                                                          self._config_inotify_workers(conf_orig)))
        self._inotify.init(conf)

        for name, inoplug in six.iteritems(INOPLUGS):
            LOG.info("inoplug init: %r", name)
            inoplug.init(conf)
            LOG.info("inoplug safe_init: %r", name)
            inoplug.safe_init()
            if inoplug.enabled and inoplug.autostart:
                LOG.info("inoplug at_start: %r", name)
                inoplug.at_start()

        self._inotify.start()

    def _pre_cache(self, toppath, srcpath, force = False):
        xid  = self._module.get_state_value('started_at').replace(' ', 'T')
        fkey = self._mk_filekey(srcpath, xid)
        nsum = self._mk_sum(srcpath)

        if not force and self._chk_file_cached(toppath, srcpath, fkey, nsum):
            return False

        return (fkey, nsum)

    def _post_cache(self, toppath, srcpath, fkey, nsum):
        with self._lock_fcached:
            if toppath not in self._files_cached:
                self._files_cached[toppath] = {}
            self._files_cached[toppath][srcpath] = nsum

        self._file_sync.set_key(fkey, nsum, self._object_ttl)

    def _upload(self, toppath, srcpath, destpath, workers = None, wpriority = None, **opts):
        if not workers:
            workers = self._wupload

        xopts = {'add_uploaded': True,
                 'cache':        True,
                 'delete':       False,
                 'max_attempts': 1}
        xopts.update(opts)

        res = None

        if xopts['cache']:
            res = self._pre_cache(toppath, srcpath)
            if not res:
                return False

        if xopts['add_uploaded']:
            with self._lock_fuploaded:
                if destpath not in self._files_uploaded:
                    self._files_uploaded[destpath] = {'toppath': toppath,
                                                      'srcpath': srcpath,
                                                      'destpath': destpath,
                                                      'wait': None}

        workers.run(self._s3up,
                    complete     = self._s3up_complete,
                    qpriority    = wpriority,
                    src          = srcpath,
                    dst          = destpath,
                    max_attempts = xopts['max_attempts'],
                    delete       = xopts['delete'])

        if xopts['cache']:
            self._post_cache(toppath, srcpath, res[0], res[1])

        return True

    def _dir_to_compress(self, toppath, params, xvars):
        r = {}

        if not os.path.isdir(toppath):
            return r

        for xdir in os.listdir(toppath):
            if self._killed:
                raise GeneratorExit()

            subdir    = os.path.join(toppath, xdir)

            if xdir in ('.', '..') or not os.path.isdir(subdir):
                continue

            if params['exclude_patterns'] and params['exclude_patterns'](subdir):
                LOG.debug("exclude path from scan. (path: %r)", subdir)
                continue

            destpath  = self._destpath(toppath,
                                       params['compress_dir']['dest_params'],
                                       subdir,
                                       xvars)
            r[subdir] = {'comppath': "%s.tar" % destpath,
                         'cnt':      0,
                         'dirpath':  subdir,
                         'size':     0}

            for root, dirs, files in os.walk(subdir, topdown = True, followlinks = params['follow']):
                stop_iter = False

                if self._killed:
                    raise GeneratorExit()

                if params['exclude_patterns'] and params['exclude_patterns'](root):
                    dirs[:] = []
                    LOG.debug("exclude path from scan. (path: %r)", root)
                    break

                if not files:
                    continue

                for filename in files:
                    srcpath = os.path.join(root, filename)

                    if params['exclude_patterns'] and params['exclude_patterns'](srcpath):
                        LOG.debug("exclude file from scan. (filepath: %r)", srcpath)
                        continue

                    filesize = os.path.getsize(srcpath)
                    if filesize >= params['compress_dir']['max_file_size']:
                        stop_iter = True
                        break

                    r[root]['size'] += filesize
                    r[root]['cnt']  += 1
                    if r[root]['size'] >= params['compress_dir']['max_total_size']:
                        stop_iter = True
                        break

                if stop_iter:
                    del r[subdir]
                    break

            if subdir not in r or r[subdir]['cnt'] == 0:
                LOG.info("database %r isn't compressible", subdir)
                r.pop(subdir, None)

        return r

    def _compress_dir(self, comppath, toppath, dirroot, destpath, params):
        LOG.info("writing tar file: %r", comppath)

        workers_upload = params['compress_dir']['workers_upload']
        max_attempts   = params['compress_dir']['max_attempts']
        tar            = None

        try:
            helpers.make_dirs(os.path.dirname(comppath))

            with tarfile.open(comppath, 'w') as tar:
                self._files_compressed.append(comppath)

                # pylint: disable-msg=unused-variable
                for root, dirs, files in os.walk(dirroot, topdown = True, followlinks = params['follow']):
                    if self._killed:
                        raise GeneratorExit()

                    for filename in files:
                        if self._killed:
                            raise GeneratorExit()

                        srcpath = os.path.join(root, filename)

                        if params['exclude_patterns'] and params['exclude_patterns'](srcpath):
                            LOG.debug("exclude file from scan. (filepath: %r)", srcpath)
                            continue

                        res = self._pre_cache(toppath, srcpath)
                        if not res:
                            continue

                        tar.add(srcpath)

                        self._post_cache(toppath, srcpath, res[0], res[1])
        except GeneratorExit:
            return (GeneratorExit(), toppath, comppath, destpath, workers_upload, max_attempts)
        except Exception as e:
            return (Exception(e), toppath, comppath, destpath, workers_upload, max_attempts)
        finally:
            if tar:
                tar.close()

        return (True, toppath, comppath, destpath, workers_upload, max_attempts)

    def _compress_dir_complete(self, ret):
        result, toppath, comppath, destpath, workers_upload, max_attempts = ret

        if isinstance(result, GeneratorExit):
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        elif isinstance(result, Exception):
            self._module.add_errors(repr(result))
            LOG.exception(result)
        elif result is True:
            if workers_upload:
                workers   = self._wcompress
                wpriority = -1
            else:
                workers   = self._wupload
                wpriority = None

            self._upload(toppath,
                         comppath,
                         destpath,
                         workers      = workers,
                         wpriority    = wpriority,
                         cache        = False,
                         delete       = True,
                         max_attempts = max_attempts)

    def _walk_from_cache(self, toppath, params, xvars):
        LOG.info("walk from cache: %r", toppath)

        with self._lock_fcached:
            if not toppath in self._files_cached:
                return
            files_cached = copy.deepcopy(self._files_cached[toppath])

        for srcpath in six.iterkeys(files_cached[toppath]):
            if self._killed:
                raise GeneratorExit()

            if params['second_phase'].get('max_attempts'):
                max_attempts = self._get_max_attempts(params['second_phase'])
            else:
                max_attempts = self._get_max_attempts(params)

            destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)
            self._upload(toppath,
                         srcpath,
                         destpath,
                         workers      = None,
                         wpriority    = None,
                         max_attempts = max_attempts)

        del files_cached

    def _walk_from_inotify(self, toppath, params, xvars):
        LOG.info("walk from inotify: %r", toppath)

        if not self._inotify:
            return False

        plug_name = BackupsSyncInoPlug.PLUGIN_NAME
        files_inotified = copy.deepcopy(INOPLUGS[plug_name].get_files(toppath))
        if not files_inotified:
            return files_inotified

        for srcpath in files_inotified:
            if self._killed:
                raise GeneratorExit()

            if params['second_phase'].get('max_attempts'):
                max_attempts = self._get_max_attempts(params['second_phase'])
            else:
                max_attempts = self._get_max_attempts(params)

            destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)
            self._upload(toppath,
                         srcpath,
                         destpath,
                         workers      = None,
                         wpriority    = None,
                         max_attempts = max_attempts)

        del files_inotified

        return True

    def _walk(self, toppath, params, xvars, dircomp = None):
        LOG.info("walk path: %r", toppath)

        if self._phase > 1 and params['second_phase'].get('max_attempts'):
            max_attempts = self._get_max_attempts(params['second_phase'])
        else:
            max_attempts = self._get_max_attempts(params)

        for root, dirs, files in os.walk(toppath, topdown = True, followlinks = params['follow']):
            if self._killed:
                raise GeneratorExit()

            if params['exclude_patterns'] and params['exclude_patterns'](root):
                dirs[:] = []
                LOG.debug("exclude path from scan. (path: %r)", root)
                continue

            if not dirs and not files:
                srcpath  = os.path.join(root, XTRARGON_FILE)
                destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)
                self._put_file(destpath)
                continue

            if self._phase == 1 and self._wcompress and dircomp and root in dircomp:
                comppath = dircomp[root]['comppath']
                srcpath  = os.path.join(dircomp[root]['dirpath'],
                                        os.path.basename(comppath))
                destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)

                self._wcompress.run(self._compress_dir,
                                    complete   = self._compress_dir_complete,
                                    qpriority  = time.time(),
                                    comppath   = comppath,
                                    toppath    = toppath,
                                    dirroot    = root,
                                    destpath   = destpath,
                                    params     = params)
                dirs[:]  = []
                files[:] = []
                continue

            for filename in files:
                if self._killed:
                    raise GeneratorExit()

                srcpath = os.path.join(root, filename)

                if params['exclude_patterns'] and params['exclude_patterns'](srcpath):
                    LOG.debug("exclude file from scan. (filepath: %r)", srcpath)
                    continue

                destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)

                self._upload(toppath,
                             srcpath,
                             destpath,
                             workers      = None,
                             wpriority    = None,
                             max_attempts = max_attempts)

    def _wait_for_next_step(self):
        LOG.info("wait for next step. (phase: %r)", self._phase)

        if self._wcompress:
            while not self._killed:
                if self._wcompress.killable():
                    break

                time.sleep(0.5)

        if self._wupload:
            while not self._killed:
                if self._wupload.killable():
                    break

                time.sleep(0.5)

    def _wait_files_uploaded(self):
        LOG.info("wait for files uploaded. (phase: %r)", self._phase)

        while True:
            files_uploaded = {}

            if self._killed:
                raise GeneratorExit()

            with self._lock_fuploaded:
                if not self._files_uploaded:
                    return

                files_uploaded = copy.deepcopy(self._files_uploaded)

            for key, item in six.iteritems(files_uploaded):
                if self._killed:
                    raise GeneratorExit()

                if not os.path.exists(item['srcpath']):
                    with self._lock_fuploaded:
                        self._files_uploaded.pop(key, None)
                    continue

                with self._lock_fuploaded:
                    if key not in self._files_uploaded \
                       or self._files_uploaded[key].get('wait'):
                        continue

                    self._files_uploaded[key]['wait'] = self._phase

                LOG.info("wait file upload. (phase: %r, file: %r)", self._phase, key)

                if self._upload(item['toppath'],
                                item['srcpath'],
                                item['destpath'],
                                workers      = None,
                                wpriority    = None,
                                add_uploaded = False):
                    continue

                with self._lock_fuploaded:
                    self._files_uploaded.pop(key, None)

            del files_uploaded

            time.sleep(0.5)

    def _wait_inotify_scan(self):
        if not self._inotify:
            return

        LOG.info("starting wait inotify scan")

        while self._inotify.is_scanning():
            if self._killed:
                raise GeneratorExit()
            time.sleep(0.5)

        LOG.info("stopping wait inotify scan")

    def _lock_db(self):
        if self._module.lock_db:
            self._module.lock_db.set()
        if self._module.db_locked:
            self._module.db_locked.wait()

    def _run(self, fkwargs):
        if self._phase == 1:
            load_inotify = False
            conf_inotify = {}

            for toppath, params in six.iteritems(self._config['backups']['paths']):
                params['second_phase'] = self._second_phase_params(params.get('second_phase'))
                if not conf_inotify:
                    conf_inotify = params['inotify']

                if params['second_phase']['enabled'] \
                   and params['second_phase']['walk_from'] == 'inotify' \
                   and toppath in conf_inotify['paths']:
                    load_inotify = True
                    self._config_inotify_path(toppath, conf_inotify)

            if load_inotify:
                self._load_inotify(conf_inotify)

        self._wait_inotify_scan()

        for toppath, params in six.iteritems(self._config['backups']['paths']):
            toppath = os.path.abspath(toppath)
            dircomp = None

            if self._killed:
                raise GeneratorExit()

            params['compress_dir'] = self._compress_dir_params(params.get('compress_dir'))
            #params['second_phase']  = self._second_phase_params(params.get('second_phase'))

            if self._phase == 1 and params['compress_dir']['enabled']:
                if not self._wcompress:
                    self._wcompress = self._load_workers('wcompress',
                                                         self._config['backups']['compress'],
                                                         _queue.PriorityQueue())
                dircomp = self._dir_to_compress(toppath, params, fkwargs)

            if self._phase == 1:
                self._walk(toppath, params, fkwargs, dircomp)
            elif not params['second_phase']['enabled']:
                continue
            elif params['second_phase']['walk_from'] == 'inotify':
                if self._walk_from_inotify(toppath, params, fkwargs) is False:
                    self._walk(toppath, params, fkwargs, dircomp)
            elif params['second_phase']['walk_from'] == 'cache' \
               and toppath in self._files_cached:
                self._walk_from_cache(toppath, params, fkwargs)
            else:
                self._walk(toppath, params, fkwargs, dircomp)

        dircomp = None

        self._wait_for_next_step()
        self._wait_files_uploaded()
        self._wait_for_next_step()

    def _set_files_errors(self):
        if not self._files_errors:
            return

        ferrors = copy.deepcopy(self._files_errors)
        for ferror in six.itervalues(ferrors):
            self._module.add_errors(ferror)

        self._files_errors = {}

    def _clean_files_compressed(self):
        if not self._files_compressed:
            return

        dirs_comp = {}

        for tar in self._files_compressed:
            dirs_comp[os.path.dirname(tar)] = 1
            if os.path.exists(tar):
                os.unlink(tar)

        for dir_comp in six.iterkeys(dirs_comp):
            try:
                os.rmdir(dir_comp)
            except OSError:
                pass

        del dirs_comp

        self._files_compressed = []

    def _post_run(self, fkwargs):
        LOG.info("finalizing backup")

        self._files_cached   = {}
        self._files_uploaded = {}

        try:
            self._set_files_errors()

            state = self._module.get_state(end = True)
            state_file_format = self._config['backups'].get('state_file_format') or DEFAULT_STATE_FILE_FORMAT
            self._put_file(state_file_format.format(**fkwargs),
                           body = json.dumps(state))

            if self._config['backups'].get('notifications'):
                self._module.notifiers(state)

            for name, inoplug in six.iteritems(INOPLUGS):
                LOG.info("inoplug at_stop: %r", name)
                inoplug.at_stop()

            if self._inotify:
                self._inotify.stop()
                self._inotify = None

            if self._wcompress:
                self._wcompress.killall(1)
                self._wcompress = None

            if self._wupload:
                self._wupload.killall(1)
                self._wupload = None

            self._clean_files_compressed()

            gc.collect()
        except Exception as e:
            LOG.exception(e)

    def run(self):
        self._module.is_running.wait()
        LOG.info("starting backup")
        self._module.set_state_value('started_at', time.strftime(DATE_FORMAT))
        self._module.set_state_value('uuid', "%s" % uuid.uuid4())
        self._module.set_state_value('status', STATUS_PROCESSING)

        if self._config['backups'].get('notifications'):
            self._module.notifiers(self._module.get_state())

        self._module.is_scanning.set()

        self._files_cached     = {}
        self._files_uploaded   = {}
        self._files_errors     = {}
        self._files_compressed = []

        fkwargs = {'env': os.environ,
                   'backup_type': self._backup_type,
                   'gmtime': datetime.utcnow(),
                   'time': datetime.now(),
                   'server_id': self._module.server_id}

        try:
            self._wupload = self._load_workers('wupload',
                                               self._config['backups']['upload'])

            if self._killed:
                raise GeneratorExit()

            LOG.info("first phase of backup")
            self._phase = 1
            self._run(fkwargs)

            if self._killed:
                raise GeneratorExit()

            self._lock_db()

            LOG.info("second phase of backup")
            self._phase = 2
            self._run(fkwargs)
        except GeneratorExit:
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        except Exception as e:
            self._module.add_errors(repr(e))
            LOG.exception(e)
        finally:
            if self._module.lock_db:
                self._module.lock_db.clear()
            self._module.is_scanning.clear()
            self._module.is_running.clear()

        self._post_run(fkwargs)

    def terminate(self):
        LOG.info("backup terminated")
        self._killed = True
        self._module.is_running.set()
        self._module.is_scanning.set()

        for name, inoplug in six.iteritems(INOPLUGS):
            LOG.info("inoplug at_stop: %r", name)
            inoplug.at_stop()

        if self._inotify:
            self._inotify.stop()

        if self._wcompress:
            self._wcompress.killall(1)

        if self._wupload:
            self._wupload.killall(1)

        self._clean_files_compressed()
